#pylint: disable=too-many-arguments
from collections import OrderedDict
from datetime import datetime, date
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element
from decimal import Decimal
import pytz
from django.core.exceptions import ValidationError
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from jutil.format import dec2
from jutil.parse import parse_datetime
from jutil.validators import iban_filter, iban_validator, iso_payment_reference_validator, \
    fi_payment_reference_validator, ascii_filter
from jutil.xml import xml_to_dict, _xml_element_set_data_r


PAIN001_REMITTANCE_INFO_MSG = 'M'
PAIN001_REMITTANCE_INFO_OCR = 'O'
PAIN001_REMITTANCE_INFO_OCR_ISO = 'I'

PAIN001_REMITTANCE_INFO_TYPE = (
    (PAIN001_REMITTANCE_INFO_MSG, _('message')),
    (PAIN001_REMITTANCE_INFO_OCR, _('OCR')),
    (PAIN001_REMITTANCE_INFO_OCR_ISO, _('OCR/ISO')),
)

PAIN001_REMITTANCE_INFO_VALUES = [t[0] for t in PAIN001_REMITTANCE_INFO_TYPE]


class Pain001Party:
    def __init__(self, name: str, account: str, bic: str, org_id: str = '', address_lines: list or None = None, country_code: str = ''):
        if address_lines is None:
            address_lines = []
        account = iban_filter(account)
        iban_validator(account)
        self.name = name
        self.account = account
        self.bic = bic
        self.org_id = org_id
        self.address_lines = address_lines
        self.country_code = country_code


class Pain001Payment:
    def __init__(self, payment_id, creditor: Pain001Party, amount: Decimal, remittance_info: str, remittance_info_type: str, due_date: date):
        self.payment_id = payment_id
        self.creditor = creditor
        self.amount = amount
        self.remittance_info = remittance_info
        self.remittance_info_type = remittance_info_type
        self.due_date = due_date

    def clean(self):
        if not self.remittance_info:
            raise ValidationError(_('pain001.remittance.info.missing'))
        if self.remittance_info_type not in PAIN001_REMITTANCE_INFO_VALUES:
            raise ValidationError(_('pain001.remittance.info.type.invalid'))
        if self.remittance_info_type == PAIN001_REMITTANCE_INFO_MSG:
            if not self.remittance_info:
                raise ValidationError(_('Invalid payment reference: {}').format(self.remittance_info))
        elif self.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR:
            fi_payment_reference_validator(self.remittance_info)
        elif self.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR_ISO:
            iso_payment_reference_validator(self.remittance_info)


class Pain001:
    """
    Class for generating pain.001.001.03 SEPA payment XML files.
    """

    pain_element_name = 'CstmrCdtTrfInitn'
    tz_str = 'Europe/Helsinki'
    tz = None

    def __init__(self, msg_id: str,
                 debtor_name: str,
                 debtor_account: str,
                 debtor_bic: str,
                 debtor_org_id: str,
                 debtor_address_lines: list,
                 debtor_country_code: str):
        self.msg_id = msg_id
        self.debtor = Pain001Party(debtor_name, debtor_account, debtor_bic, debtor_org_id, debtor_address_lines, debtor_country_code)
        self.payments = []

    def add_payment(self, payment_id,
                    creditor_name: str,
                    creditor_account: str,
                    creditor_bic: str,
                    amount: Decimal,
                    remittance_info: str,
                    remittance_info_type: str = PAIN001_REMITTANCE_INFO_MSG,
                    due_date: date = None,
                    ):
        if not due_date:
            due_date = self._local_time().date()
        creditor = Pain001Party(creditor_name, creditor_account, creditor_bic)
        p = Pain001Payment(payment_id, creditor, dec2(amount), remittance_info, remittance_info_type, due_date)
        p.clean()
        self.payments.append(p)

    def _ctrl_sum(self) -> Decimal:
        total = Decimal('0.00')
        for p in self.payments:
            assert isinstance(p, Pain001Payment)
            total += p.amount
        return total

    def _append_simple(self, parent: Element, tag: str, value):
        e = Element(tag)
        e.text = str(value)
        parent.append(e)
        return e

    def _local_time(self, t: datetime or None = None) -> datetime:
        if not t:
            t = now()
        if not self.tz:
            self.tz = pytz.timezone(self.tz_str)
        return t.astimezone(self.tz)

    def _timestamp(self, t: datetime) -> str:
        return self._local_time(t).isoformat()

    def _dict_to_element(self, doc: dict, value_key: str = '@', attribute_prefix: str = '@') -> Element:
        if len(doc) != 1:
            raise Exception('Invalid data dict for XML generation, document root must have single element')
        for tag, data in doc.items():
            el = Element(tag)
            assert isinstance(el, Element)
            _xml_element_set_data_r(el, data, value_key, attribute_prefix)
            return el

    def _grp_hdr(self) -> Element:
        g = Element('GrpHdr')
        self._append_simple(g, 'MsgId', self.msg_id)
        self._append_simple(g, 'CreDtTm', self._timestamp(now()))
        self._append_simple(g, 'NbOfTxs', len(self.payments))
        self._append_simple(g, 'CtrlSum', self._ctrl_sum())
        # self._append_simple(g, 'BtchBookg', 'true')  # debit all at once
        # self._append_simple(g, 'Grpg', 'MIXD')
        g.append(self._dict_to_element({
            'InitgPty': OrderedDict([
                ('Nm', self.debtor.name),
                ('PstlAdr', OrderedDict([
                    ('Ctry', self.debtor.country_code),
                    ('AdrLine', [{'@': l} for l in self.debtor.address_lines]),
                ])),
            ]),
        }))
        return g

    def _pmt_inf(self, p: Pain001Payment) -> Element:
        if p.remittance_info_type == PAIN001_REMITTANCE_INFO_MSG:
            rmt_inf = ['RmtInf', OrderedDict([
                ('Ustrd', p.remittance_info),
            ])]
        elif p.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR:
            rmt_inf = ['RmtInf', OrderedDict([
                ('Strd', OrderedDict([
                    ('CdtrRefInf', OrderedDict([
                        ('Tp', OrderedDict([
                            ('CdOrPrtry', OrderedDict([
                                ('Cd', 'SCOR'),
                            ])),
                        ])),
                        ('Ref', p.remittance_info),
                    ])),
                ])),
            ])]
        elif p.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR_ISO:
            rmt_inf = ['RmtInf', OrderedDict([
                ('Strd', OrderedDict([
                    ('CdtrRefInf', OrderedDict([
                        ('Tp', OrderedDict([
                            ('CdOrPrtry', OrderedDict([
                                ('Cd', 'SCOR'),
                            ])),
                            ('Issr', 'ISO'),
                        ])),
                        ('Ref', p.remittance_info),
                    ])),
                ])),
            ])]
        else:
            raise ValidationError(_('Invalid remittance info type: {}').format(p.remittance_info_type))

        return self._dict_to_element({
            'PmtInf': OrderedDict([
                ('PmtInfId', str(p.payment_id)),
                ('PmtMtd', 'TRF'),  # payment order
                ('ReqdExctnDt', p.due_date.isoformat()),
                ('Dbtr', OrderedDict([
                    ('Nm', self.debtor.name),
                    ('PstlAdr', OrderedDict([
                        ('Ctry', self.debtor.country_code),
                        ('AdrLine', [{'@': l} for l in self.debtor.address_lines]),
                    ])),
                    ('Id', OrderedDict([
                        ('OrgId', OrderedDict([
                            ('Othr', OrderedDict([
                                ('Id', self.debtor.org_id),
                                ('SchmeNm', OrderedDict([
                                    ('Cd', 'BANK'),
                                ])),
                            ])),
                        ])),
                    ])),
                ])),
                ('DbtrAcct', OrderedDict([
                    ('Id', OrderedDict([
                        ('IBAN', self.debtor.account),
                    ])),
                ])),
                ('DbtrAgt', OrderedDict([
                    ('FinInstnId', OrderedDict([
                        ('BIC', self.debtor.bic),
                    ])),
                ])),
                ('ChrgBr', 'SLEV'),  # FollowingService level
                ('CdtTrfTxInf', OrderedDict([
                    ('PmtId', OrderedDict([
                        ('EndToEndId', str(p.payment_id)),
                    ])),
                    ('Amt', OrderedDict([
                        ('InstdAmt', {'@': str(p.amount), '@Ccy': 'EUR'}),
                    ])),
                    ('UltmtDbtr', OrderedDict([
                        ('Nm', self.debtor.name),
                    ])),
                    ('CdtrAgt', OrderedDict([
                        ('FinInstnId', OrderedDict([
                            ('BIC', p.creditor.bic),
                        ])),
                    ])),
                    ('Cdtr', OrderedDict([
                        ('Nm', ascii_filter(p.creditor.name)),
                    ])),
                    ('CdtrAcct', OrderedDict([
                        ('Id', OrderedDict([
                            ('IBAN', p.creditor.account),
                        ])),
                    ])),
                    rmt_inf,
                ])),
            ]),
        })

    def render_to_element(self) -> Element:
        if not self.payments:
            raise ValidationError('No payments in pain.001.001.03')
        doc = Element('Document', xmlns="urn:iso:std:iso:20022:tech:xsd:pain.001.001.03")
        pain = Element(self.pain_element_name)
        doc.append(pain)
        pain.append(self._grp_hdr())
        for p in self.payments:
            assert isinstance(p, Pain001Payment)
            pain.append(self._pmt_inf(p))
        return doc

    def render_to_bytes(self, doc: Element or None = None) -> bytes:
        doc = doc or self.render_to_element()
        xml_bytes = ET.tostring(doc, encoding='utf-8', method='xml')
        return xml_bytes

    def render_to_file(self, filename: str, xml_bytes: bytes or None = None):
        xml_bytes = xml_bytes or self.render_to_bytes()
        with open(filename, 'wb') as fp:
            fp.write(xml_bytes)


class Pain002:
    """
    Class for parsing pain.002.001.03 SEPA payment status XML files.
    """
    credit_datetime = None
    msg_id = None
    original_msg_id = None
    group_status = None
    status_reason = None

    def __init__(self, file_content: bytes):
        self.data = xml_to_dict(file_content)

        rpt = self.data.get('CstmrPmtStsRpt', {})

        grp_hdr = rpt.get('GrpHdr', {})
        credit_datetime_str = grp_hdr.get('CreDtTm')
        self.credit_datetime = parse_datetime(credit_datetime_str)
        self.msg_id = grp_hdr.get('MsgId')

        grp_inf = rpt.get('OrgnlGrpInfAndSts', {})
        self.original_msg_id = grp_inf.get('OrgnlMsgId')
        self.group_status = grp_inf.get('GrpSts')
        self.status_reason = grp_inf.get('StsRsnInf', {}).get('Rsn', {}).get('Prtry', '')

        if not self.msg_id:
            raise ValidationError('MsgId missing')
        if not self.original_msg_id:
            raise ValidationError('OrgnlMsgId missing')
        if not self.group_status:
            raise ValidationError('GrpSts missing')

    def __str__(self):
        return '{}: {} {} {}'.format(self.msg_id, self.original_msg_id, self.group_status, self.status_reason)

    @property
    def is_accepted(self):
        return self.group_status == 'ACCP'

    @property
    def is_rejected(self):
        return self.group_status == 'RJCT'



# Generated by Django 2.2.3 on 2019-11-28 22:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0045_wsediconnection_receiver_identifier'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='wsedisoapcall',
            options={'verbose_name': 'WS-EDI SOAP call', 'verbose_name_plural': 'WS-EDI SOAP calls'},
        ),
    ]



import logging
import pytz
from django.core.management import CommandParser
from jutil.command import SafeCommand
from jbank.helpers import parse_start_and_end_date
from jbank.models import WsEdiConnection
from jbank.wsedi import wsedi_execute


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
    Executes WS-EDI command using direct bank connection.
    """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--ws', type=int, default=1)
        parser.add_argument('--cmd', type=str, default='DownloadFileList')
        parser.add_argument('--file-reference', type=str)
        parser.add_argument('--file-type', type=str)
        parser.add_argument('--start-date', type=str)
        parser.add_argument('--end-date', type=str)

    def do(self, *args, **options):
        ws = WsEdiConnection.objects.get(id=options['ws'])
        assert isinstance(ws, WsEdiConnection)
        if ws and not ws.enabled:
            logger.info('WS connection %s not enabled, exiting', ws)
            return

        start_date, end_date = parse_start_and_end_date(pytz.timezone('Europe/Helsinki'), **options)
        cmd = options['cmd']
        file_reference = options['file_reference']
        file_type = options['file_type']
        wsedi_execute(ws, command=cmd, file_reference=file_reference, file_type=file_type, start_date=start_date, end_date=end_date, verbose=True)



import logging
from datetime import datetime, date
from os.path import basename
import pytz
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from jacc.models import Account, AccountType, EntryType
from jbank.models import Statement, StatementRecord, StatementRecordSepaInfo, ReferencePaymentBatch, \
    ReferencePaymentRecord, StatementFile, ReferencePaymentBatchFile, Payout, PayoutStatus, PAYOUT_PAID
from jbank.sepa import Pain002
import re
from lxml import etree, objectify
from jutil.parse import parse_datetime


ASSIGNABLE_STATEMENT_HEADER_FIELDS = (
    'account_number',
    'statement_number',
    'begin_date',
    'end_date',
    'record_date',
    'customer_identifier',
    'begin_balance_date',
    'begin_balance',
    'record_count',
    'currency_code',
    'account_name',
    'account_limit',
    'owner_name',
    'contact_info_1',
    'contact_info_2',
    'bank_specific_info_1',
    'iban',
    'bic',
)

ASSIGNABLE_STATEMENT_RECORD_FIELDS = (
    'record_number',
    'archive_identifier',
    'record_date',
    'value_date',
    'paid_date',
    'entry_type',
    'record_code',
    'record_description',
    'amount',
    'receipt_code',
    'delivery_method',
    'name',
    'name_source',
    'recipient_account_number',
    'recipient_account_number_changed',
    'remittance_info',
)

ASSIGNABLE_STATEMENT_RECORD_SEPA_INFO_FIELDS = (
    'reference',
    'iban_account_number',
    'bic_code',
    'recipient_name_detail',
    'payer_name_detail',
    'identifier',
    'archive_identifier',
)

ASSIGNABLE_REFERENCE_PAYMENT_BATCH_HEADER_FIELDS = (
    'record_date',
    'institution_identifier',
    'service_identifier',
    'currency_identifier',
)

ASSIGNABLE_REFERENCE_PAYMENT_RECORD_FIELDS = (
    'record_type',
    'account_number',
    'record_date',
    'paid_date',
    'archive_identifier',
    'remittance_info',
    'payer_name',
    'currency_identifier',
    'name_source',
    'amount',
    'correction_identifier',
    'delivery_method',
    'receipt_code',
)


logger = logging.getLogger(__name__)


@transaction.atomic  # noqa
def create_statement(statement_data: dict, name: str, file: StatementFile, **kw) -> Statement:
    """
    Creates Statement from statement data parsed by parse_tiliote_statements()
    :param statement_data: See parse_tiliote_statements
    :param name: File name of the account statement
    :param file: Source statement file
    :return: Statement
    """
    #pylint: disable=too-many-locals,too-many-branches
    if 'header' not in statement_data or not statement_data['header']:
        raise ValidationError('Invalid header field in statement data {}: {}'.format(name, statement_data.get('header')))
    header = statement_data['header']

    account_number = header['account_number']
    if not account_number:
        raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=''))
    accounts = list(Account.objects.filter(name=account_number))
    if len(accounts) != 1:
        raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=account_number))
    account = accounts[0]
    assert isinstance(account, Account)

    if Statement.objects.filter(name=name, account=account).first():
        raise ValidationError('Bank account {} statement {} of processed already'.format(account_number, name))
    stm = Statement(name=name, account=account, file=file)
    for k in ASSIGNABLE_STATEMENT_HEADER_FIELDS:
        if k in header:
            setattr(stm, k, header[k])
    # pprint(statement_data['header'])
    for k, v in kw.items():
        setattr(stm, k, v)
    stm.full_clean()
    stm.save()

    if EntryType.objects.filter(code=settings.E_BANK_DEPOSIT).count() == 0:
        raise ValidationError(_('entry.type.missing') + ' ({}): {}'.format('settings.E_BANK_DEPOSIT', settings.E_BANK_DEPOSIT))
    if EntryType.objects.filter(code=settings.E_BANK_WITHDRAW).count() == 0:
        raise ValidationError(_('entry.type.missing') + ' ({}): {}'.format('settings.E_BANK_WITHDRAW', settings.E_BANK_WITHDRAW))
    entry_types = {
        '1': EntryType.objects.get(code=settings.E_BANK_DEPOSIT),
        '2': EntryType.objects.get(code=settings.E_BANK_WITHDRAW),
    }

    for rec_data in statement_data['records']:
        line_number = rec_data['line_number']
        e_type = entry_types.get(rec_data['entry_type'])
        rec = StatementRecord(statement=stm, account=account, type=e_type, line_number=line_number)
        for k in ASSIGNABLE_STATEMENT_RECORD_FIELDS:
            if k in rec_data:
                setattr(rec, k, rec_data[k])
        for k in ('messages', 'client_messages', 'bank_messages'):
            if k in rec_data:
                setattr(rec, k, '\n'.join(rec_data[k]))
        rec.full_clean()
        rec.save()

        if 'sepa' in rec_data:
            sepa_info_data = rec_data['sepa']
            sepa_info = StatementRecordSepaInfo(record=rec)
            for k in ASSIGNABLE_STATEMENT_RECORD_SEPA_INFO_FIELDS:
                if k in sepa_info_data:
                    setattr(sepa_info, k, sepa_info_data[k])
            # pprint(rec_data['sepa'])
            sepa_info.full_clean()
            sepa_info.save()
    # pylint: enable=too-many-locals,too-many-branches


@transaction.atomic
def create_reference_payment_batch(batch_data: dict, name: str, file: ReferencePaymentBatchFile, **kw) -> ReferencePaymentBatch:
    """
    Creates ReferencePaymentBatch from data parsed by parse_svm_batches()
    :param batch_data: See parse_svm_batches
    :param name: File name of the batch file
    :return: ReferencePaymentBatch
    """
    if ReferencePaymentBatch.objects.exclude(file=file).filter(name=name).first():
        raise ValidationError('Reference payment batch file {} already exists'.format(name))

    if 'header' not in batch_data or not batch_data['header']:
        raise ValidationError('Invalid header field in reference payment batch data {}: {}'.format(name, batch_data.get('header')))
    header = batch_data['header']

    batch = ReferencePaymentBatch(name=name, file=file)
    for k in ASSIGNABLE_REFERENCE_PAYMENT_BATCH_HEADER_FIELDS:
        if k in header:
            setattr(batch, k, header[k])
    # pprint(statement_data['header'])
    for k, v in kw.items():
        setattr(batch, k, v)
    batch.full_clean()
    batch.save()
    e_type = EntryType.objects.get(code=settings.E_BANK_REFERENCE_PAYMENT)

    for rec_data in batch_data['records']:
        line_number = rec_data['line_number']
        account_number = rec_data['account_number']
        if not account_number:
            raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=''))
        accounts = list(Account.objects.filter(name=account_number))
        if len(accounts) != 1:
            raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=account_number))
        account = accounts[0]
        assert isinstance(account, Account)

        rec = ReferencePaymentRecord(batch=batch, account=account, type=e_type, line_number=line_number)
        for k in ASSIGNABLE_REFERENCE_PAYMENT_RECORD_FIELDS:
            if k in rec_data:
                setattr(rec, k, rec_data[k])
        # pprint(rec_data)
        rec.full_clean()
        rec.save()


def get_or_create_bank_account(account_number: str, currency: str = 'EUR') -> Account:
    a_type = AccountType.objects.get_or_create(code=settings.ACCOUNT_BANK_ACCOUNT, is_asset=True, defaults={'name': _('bank account')})[0]
    acc = Account.objects.get_or_create(name=account_number, type=a_type, currency=currency)[0]
    return acc


def process_pain002_file_content(bcontent: bytes, filename: str, created: datetime or None = None):
    if not created:
        created = now()
    s = Pain002(bcontent)
    p = Payout.objects.filter(msg_id=s.original_msg_id).first()
    ps = PayoutStatus(payout=p, file_name=basename(filename), msg_id=s.msg_id, original_msg_id=s.original_msg_id,
                      group_status=s.group_status, status_reason=s.status_reason[:255], created=created)
    ps.full_clean()
    fields = (
        'payout',
        'file_name',
        'response_code',
        'response_text',
        'msg_id',
        'original_msg_id',
        'group_status',
        'status_reason',
    )
    params = {}
    for k in fields:
        params[k] = getattr(ps, k)
    ps_old = PayoutStatus.objects.filter(**params).first()
    if ps_old:
        ps = ps_old
    else:
        ps.save()
        logger.info('%s status updated %s', p, ps)
    if p:
        if ps.is_accepted:
            p.state = PAYOUT_PAID
            p.paid_date = s.credit_datetime
            p.save(update_fields=['state', 'paid_date'])
            logger.info('%s marked as paid %s', p, ps)
    return ps


def make_msg_id() -> str:
    return re.sub(r'[^\d]', '', now().isoformat())[:-4]


def validate_xml(content: bytes, xsd_file_name: str):
    """
    Validates XML using XSD
    """
    schema = etree.XMLSchema(file=xsd_file_name)
    parser = objectify.makeparser(schema=schema)
    objectify.fromstring(content, parser)


def parse_start_and_end_date(tz: pytz.tzinfo or None = None, **options) -> (date or None, date or None):
    start_date = None
    end_date = None
    time_now = now().astimezone(tz if tz else pytz.utc)
    if options['start_date']:
        if options['start_date'] == 'today':
            start_date = time_now.date()
        else:
            start_date = parse_datetime(options['start_date']).date()
        end_date = start_date
    if options['end_date']:
        if options['end_date'] == 'today':
            end_date = time_now.date()
        else:
            end_date = parse_datetime(options['end_date']).date()
    return start_date, end_date



# Generated by Django 2.2.3 on 2019-11-30 20:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0053_wsediconnection_target_identifier'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='environment',
            field=models.CharField(default='PRODUCTION', max_length=32, verbose_name='environment'),
        ),
    ]



#pylint: disable=logging-format-interpolation,too-many-locals,too-many-branches
import logging
import traceback
from django.core.management import CommandParser
from jutil.xml import xml_to_dict
from jbank.models import Payout, PayoutStatus, PAYOUT_ERROR, PAYOUT_WAITING_UPLOAD, PAYOUT_UPLOADED, WsEdiConnection
from jbank.wsedi import wsedi_upload_file, wsedi_execute
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
        Upload Finnish bank files
        """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--payout', type=int)
        parser.add_argument('--file-type', type=str, help='E.g. XL, NDCORPAYS, pain.001.001.03')
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--force', action='store_true')
        parser.add_argument('--default-ws', type=int)
        parser.add_argument('--ws', type=int)

    def do(self, *args, **options):
        default_ws = WsEdiConnection.objects.get(id=options['default_ws']) if options['default_ws'] else None
        assert default_ws is None or isinstance(default_ws, WsEdiConnection)
        file_type = options['file_type']
        if not file_type:
            print('--file-type required (e.g. XL, NDCORPAYS, pain.001.001.03)')
            return

        payouts = Payout.objects.all()
        if options['payout']:
            payouts = Payout.objects.filter(id=options['payout'])
        else:
            payouts = payouts.filter(state=PAYOUT_WAITING_UPLOAD)
            if options['ws']:
                payouts = payouts.filter(connection_id=options['ws'])

        for p in list(payouts):
            assert isinstance(p, Payout)
            p.refresh_from_db()
            ws_connection = p.connection or default_ws

            if p.state != PAYOUT_WAITING_UPLOAD:
                logger.info('Skipping {} since not in state PAYOUT_WAITING_UPLOAD'.format(p))
                continue
            if ws_connection and not ws_connection.enabled:
                logger.info('WS connection %s not enabled, skipping payment %s', ws_connection, p)
                continue

            response_code = ''
            response_text = ''
            try:
                # upload file
                logger.info('Uploading payment id={} {} file {}'.format(p.id, file_type, p.full_path))
                with open(p.full_path, 'rt') as fp:
                    file_content = fp.read()
                p.state = PAYOUT_UPLOADED
                p.save(update_fields=['state'])
                if ws_connection:
                    content = wsedi_execute(ws_connection, 'UploadFile', file_content=file_content, file_type=file_type, verbose=options['verbose'])
                    data = xml_to_dict(content, array_tags=['FileDescriptor'])
                else:
                    res = wsedi_upload_file(file_content=file_content, file_type=file_type, file_name=p.file_name, verbose=options['verbose'])
                    logger.info('HTTP response {}'.format(res.status_code))
                    logger.info(res.text)
                    data = res.json()

                # parse response
                response_code = data.get('ResponseCode', '')[:4]
                response_text = data.get('ResponseText', '')[:255]
                if response_code != '00':
                    msg = 'WS-EDI file {} upload failed: {} ({})'.format(p.file_name, response_text, response_code)
                    logger.error(msg)
                    raise Exception('Response code {} ({})'.format(response_code, response_text))
                if 'FileDescriptors' in data:
                    fds = data.get("FileDescriptors", {}).get("FileDescriptor", [])
                    fd = {} if not fds else fds[0]
                    file_reference = fd.get('FileReference', '')
                    if file_reference:
                        p.file_reference = file_reference
                        p.save(update_fields=['file_reference'])
                PayoutStatus.objects.create(payout=p, msg_id=p.msg_id, file_name=p.file_name, response_code=response_code,
                                            response_text=response_text, status_reason='File upload OK')

            except Exception as e:
                long_err = "File upload failed ({}): ".format(p.file_name) + traceback.format_exc()
                logger.error(long_err)
                short_err = 'File upload failed: ' + str(e)
                p.state = PAYOUT_ERROR
                p.save(update_fields=['state'])
                PayoutStatus.objects.create(payout=p, msg_id=p.msg_id, file_name=p.file_name, response_code=response_code,
                                            response_text=response_text, status_reason=short_err[:255])



from datetime import datetime, date
from decimal import Decimal
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.dateparse import parse_date
from django.utils.translation import gettext as _
from jacc.models import Account, EntryType
from jutil.format import dec2, dec4
from jutil.parse import parse_datetime
from jbank.models import StatementFile, Statement, StatementRecord, DELIVERY_FROM_BANK_SYSTEM, \
    StatementRecordDetail, CurrencyExchange, StatementRecordRemittanceInfo, CurrencyExchangeSource
from jbank.parsers import parse_filename_suffix
from jutil.xml import xml_to_dict


CAMT053_STATEMENT_SUFFIXES = ('XML', 'XT', 'CAMT')

CAMT053_ARRAY_TAGS = ['Bal', 'Ntry', 'NtryDtls', 'TxDtls', 'Strd']

CAMT053_INT_TAGS = ['NbOfNtries', 'NbOfTxs']


def camt053_get_iban(data: dict) -> str:
    return data.get('BkToCstmrStmt', {}).get('Stmt', {}).get('Acct', {}).get('Id', {}).get('IBAN', '')


def camt053_get_val(data: dict, key: str, default='', required: bool = True, name: str = ''):
    if key not in data:
        if required:
            raise ValidationError(_('camt.053 field {} missing').format(name if name else key))
        return default
    return data[key]


def camt053_get_currency(data: dict, key: str, required: bool = True, name: str = ''):
    try:
        v = camt053_get_val(data, key, default=None, required=False, name=name)
        if v:
            amount = dec2(v['@'])
            currency_code = v['@Ccy']
            return amount, currency_code
    except Exception:
        pass
    if required:
        raise ValidationError(_('camt.053 field {} type {} missing or invalid').format(name, 'currency'))
    return None, ''


def camt053_get_dt(data: dict, key: str, default: datetime or None = None, required: bool = True, name: str = '') -> datetime or None:
    s = camt053_get_val(data, key, None, required, name)
    return parse_datetime(s) if s is not None else default


def camt053_get_int(data: dict, key: str, default: int or None = None, required: bool = True, name: str = '') -> int:
    s = camt053_get_val(data, key, default, required, name)
    try:
        return int(s)
    except Exception:
        pass
    raise ValidationError(_('camt.053 field {} type {} missing or invalid').format(name, 'int'))


def camt053_get_date(data: dict, key: str, default: date or None = None, required: bool = True, name: str = '') -> date:
    s = camt053_get_val(data, key, default, required, name)
    try:
        return parse_date(s)
    except Exception:
        pass
    raise ValidationError(_('camt.053 field {} type {} missing or invalid').format(name, 'date'))


def camt053_parse_statement_from_file(filename: str) -> dict:
    if parse_filename_suffix(filename).upper() not in CAMT053_STATEMENT_SUFFIXES:
        raise ValidationError(_('File {filename} has unrecognized ({suffixes}) suffix for file type "{file_type}"').format(
            filename=filename, suffixes=', '.join(CAMT053_STATEMENT_SUFFIXES), file_type='camt.053'))
    with open(filename, 'rb') as fp:
        data = xml_to_dict(fp.read(), array_tags=CAMT053_ARRAY_TAGS, int_tags=CAMT053_INT_TAGS)
        return data


def camt053_get_stmt_bal(d_stmt: dict, bal_type: str) -> (Decimal, date):
    for bal in d_stmt.get('Bal', []):
        if bal.get('Tp', {}).get('CdOrPrtry', {}).get('Cd', '') == bal_type:
            amt = Decimal(bal.get('Amt', {}).get('@', ''))
            dt = camt053_get_date(bal.get('Dt', {}), 'Dt', name='Stmt.Bal[{}].Dt.Dt'.format(bal_type))
            return amt, dt
    raise ValidationError(_('camt.053 field {} type {} missing or invalid').format('Stmt.Bal.Tp.CdOrPrty.Cd', bal_type))


def camt053_domain_from_record_code(record_domain: str) -> str:
    if record_domain == 'PMNT':
        return '700'
    if record_domain == 'LDAS':
        return '761'
    return ''


def camt053_get_unified_val(qs, k: str, default) -> str:
    v = default
    for e in qs:
        v2 = getattr(e, k)
        if v == default:
            v = v2
        elif v and v2 and v2 != v:
            return default
    return v


def camt053_get_unified_str(qs, k: str) -> str:
    return camt053_get_unified_val(qs, k, '')


@transaction.atomic  # noqa
def camt053_create_statement(statement_data: dict, name: str, file: StatementFile, **kw) -> Statement:
    """
    Creates camt.053 Statement from statement data parsed by camt053_parse_statement_from_file()
    :param statement_data: XML data in form of dict
    :param name: File name of the account statement
    :param file: Source statement file
    :return: Statement
    """
    #pylint: disable=too-many-locals,too-many-statements,too-many-branches
    account_number = camt053_get_iban(statement_data)
    if not account_number:
        raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=''))
    accounts = list(Account.objects.filter(name=account_number))
    if len(accounts) != 1:
        raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=account_number))
    account = accounts[0]
    assert isinstance(account, Account)

    d_stmt = statement_data.get('BkToCstmrStmt', {}).get('Stmt', {})
    d_acct = d_stmt.get('Acct', {})
    d_ownr = d_acct.get('Ownr', {})
    d_ntry = d_stmt.get('Ntry', [])
    d_frto = d_stmt.get('FrToDt', {})
    d_txsummary = d_stmt.get('TxsSummry', {})

    if Statement.objects.filter(name=name, account=account).first():
        raise ValidationError('Bank account {} statement {} of processed already'.format(account_number, name))
    stm = Statement(name=name, account=account, file=file)
    stm.account_number = stm.iban = account_number
    stm.bic = camt053_get_val(d_acct.get('Svcr', {}).get('FinInstnId', {}), 'BIC', name='Stmt.Acct.Svcr.FinInstnId.BIC')
    stm.statement_identifier = camt053_get_val(d_stmt, 'Id', name='Stmt.Id')
    stm.statement_number = camt053_get_val(d_stmt, 'LglSeqNb', name='Stmt.LglSeqNb')
    stm.record_date = camt053_get_dt(d_stmt, 'CreDtTm', name='Stmt.CreDtTm')
    stm.begin_date = camt053_get_dt(d_frto, 'FrDtTm', name='Stmt.FrDtTm').date()
    stm.end_date = camt053_get_dt(d_frto, 'ToDtTm', name='Stmt.ToDtTm').date()
    stm.currency_code = camt053_get_val(d_acct, 'Ccy', name='Stmt.Acct.Ccy')
    if stm.currency_code != account.currency:
        raise ValidationError(_('Account currency {account_currency} does not match statement entry currency {statement_currency}'.format(
            statement_currency=stm.currency_code, account_currency=account.currency)))
    stm.owner_name = camt053_get_val(d_ownr, 'Nm', name='Stm.Acct.Ownr.Nm')
    stm.begin_balance, stm.begin_balance_date = camt053_get_stmt_bal(d_stmt, 'OPBD')
    stm.record_count = camt053_get_int(d_txsummary.get('TtlNtries', {}), 'NbOfNtries', name='Stmt.TxsSummry.TtlNtries.NbOfNtries')
    stm.bank_specific_info_1 = camt053_get_val(d_stmt, 'AddtlStmtInf', required=False)[:1024]
    for k, v in kw.items():
        setattr(stm, k, v)
    stm.full_clean()
    stm.save()

    e_deposit = EntryType.objects.filter(code=settings.E_BANK_DEPOSIT).first()
    if not e_deposit:
        raise ValidationError(_('entry.type.missing') + ' ({}): {}'.format('settings.E_BANK_DEPOSIT', settings.E_BANK_DEPOSIT))
    assert isinstance(e_deposit, EntryType)
    e_withdraw = EntryType.objects.filter(code=settings.E_BANK_WITHDRAW).first()
    if not e_withdraw:
        raise ValidationError(_('entry.type.missing') + ' ({}): {}'.format('settings.E_BANK_WITHDRAW', settings.E_BANK_WITHDRAW))
    assert isinstance(e_withdraw, EntryType)
    e_types = {
        'CRDT': e_deposit,
        'DBIT': e_withdraw,
    }
    record_type_map = {
        'CRDT': '1',
        'DBIT': '2',
    }

    for ntry in d_ntry:
        archive_id = ntry.get('AcctSvcrRef', '')
        amount, cur = camt053_get_currency(ntry, 'Amt', name='Stmt.Ntry[{}].Amt'.format(archive_id))
        if cur != account.currency:
            raise ValidationError(_('Account currency {account_currency} does not match statement entry currency {statement_currency}'.format(
                statement_currency=cur, account_currency=account.currency)))

        cdt_dbt_ind = ntry['CdtDbtInd']
        e_type = e_types.get(cdt_dbt_ind, None)
        if not e_type:
            raise ValidationError(_('Statement entry type {} not supported').format(cdt_dbt_ind))

        rec = StatementRecord(statement=stm, account=account, type=e_type)
        rec.amount = amount
        rec.archive_identifier = archive_id
        rec.entry_type = record_type_map[cdt_dbt_ind]
        rec.record_date = record_date = camt053_get_date(ntry.get('BookgDt', {}), 'Dt', name='Stmt.Ntry[{}].BkkgDt.Dt'.format(archive_id))
        rec.value_date = camt053_get_date(ntry.get('ValDt', {}), 'Dt', name='Stmt.Ntry[{}].ValDt.Dt'.format(archive_id))
        rec.delivery_method = DELIVERY_FROM_BANK_SYSTEM

        d_bktxcd = ntry.get('BkTxCd', {})
        d_domn = d_bktxcd.get('Domn', {})
        d_family = d_domn.get('Fmly', {})
        d_prtry = d_bktxcd.get('Prtry', {})
        rec.record_domain = record_domain = camt053_get_val(d_domn, 'Cd', name='Stmt.Ntry[{}].BkTxCd.Domn.Cd'.format(archive_id))
        rec.record_code = camt053_domain_from_record_code(record_domain)
        rec.family_code = camt053_get_val(d_family, 'Cd', name='Stmt.Ntry[{}].BkTxCd.Domn.Family.Cd'.format(archive_id))
        rec.sub_family_code = camt053_get_val(d_family, 'SubFmlyCd', name='Stmt.Ntry[{}].BkTxCd.Domn.Family.SubFmlyCd'.format(archive_id))
        rec.record_description = camt053_get_val(d_prtry, 'Cd', required=False)

        rec.full_clean()
        rec.save()

        for dtl_batch in ntry.get('NtryDtls', []):
            batch_identifier = dtl_batch.get('Btch', {}).get('MsgId', '')
            dtl_ix = 0
            for dtl in dtl_batch.get('TxDtls', []):
                d = StatementRecordDetail(record=rec, batch_identifier=batch_identifier)

                d_amt_dtl = dtl.get('AmtDtls', {})
                d_txamt = d_amt_dtl.get('TxAmt', {})
                d_xchg = d_txamt.get('CcyXchg', None)

                d.amount, d.currency_code = camt053_get_currency(d_txamt, 'Amt', required=False)
                d.instructed_amount, source_currency = camt053_get_currency(d_amt_dtl.get('InstdAmt', {}), 'Amt', required=False)
                if (not d_xchg and source_currency and source_currency != d.currency_code) or (d_xchg and not source_currency):
                    raise ValidationError(_('Inconsistent Stmt.Ntry[{}].NtryDtls.TxDtls[{}].AmtDtls'.format(archive_id, dtl_ix)))

                if source_currency and source_currency != d.currency_code:
                    source_currency = camt053_get_val(d_xchg, 'SrcCcy', default=source_currency, required=False)
                    target_currency = camt053_get_val(d_xchg, 'TrgCcy', default=d.currency_code, required=False)
                    unit_currency = camt053_get_val(d_xchg, 'UnitCcy', default='', required=False)
                    exchange_rate_str = camt053_get_val(d_xchg, 'XchgRate', default='', required=False)
                    exchange_rate = dec4(exchange_rate_str) if exchange_rate_str else None
                    exchange_source = CurrencyExchangeSource.objects.get_or_create(name=account_number)[0]
                    d.exchange = CurrencyExchange.objects.get_or_create(record_date=record_date, source_currency=source_currency,
                                                                        target_currency=target_currency, unit_currency=unit_currency,
                                                                        exchange_rate=exchange_rate, source=exchange_source)[0]

                d_refs = dtl.get('Refs', {})
                d.archive_identifier = d_refs.get('AcctSvcrRef', '')
                d.end_to_end_identifier = d_refs.get('EndToEndId', '')

                d_parties = dtl.get('RltdPties', {})
                d_dbt = d_parties.get('Dbtr', {})
                d.debtor_name = d_dbt.get('Nm', '')
                d_udbt = d_parties.get('UltmtDbtr', {})
                d.ultimate_debtor_name = d_udbt.get('Nm', '')
                d_cdtr = d_parties.get('Cdtr', {})
                d.creditor_name = d_cdtr.get('Nm', '')
                d_cdtr_acct = d_parties.get('CdtrAcct', {})
                d.creditor_account = d_cdtr_acct.get('Id', {}).get('IBAN', '')

                d_rmt = dtl.get('RmtInf', {})
                d.unstructured_remittance_info = d_rmt.get('Ustrd', '')

                d_rltd_dts = dtl.get('RltdDts', {})
                d.paid_date = camt053_get_dt(d_rltd_dts, 'AccptncDtTm', required=False)

                d.full_clean()
                d.save()

                st = StatementRecordRemittanceInfo(detail=d)
                for strd in d_rmt.get('Strd', []):
                    additional_info = strd.get('AddtlRmtInf', '')
                    amount, currency_code = camt053_get_currency(strd.get('RfrdDocAmt', {}), 'RmtdAmt', required=False)
                    reference = strd.get('CdtrRefInf', {}).get('Ref', '')

                    # check if new remittance info record is needed
                    #pylint: disable=too-many-boolean-expressions
                    if additional_info and st.additional_info or amount and st.amount or reference and st.reference:
                        st = StatementRecordRemittanceInfo(detail=d)
                    #pylint: enable=too-many-boolean-expressions

                    if additional_info:
                        st.additional_info = additional_info
                    if amount:
                        st.amount, st.currency_code = amount, currency_code
                    if reference:
                        st.reference = reference

                    st.full_clean()
                    st.save()

                dtl_ix += 1

        # fill record name from details
        if not rec.name:
            if rec.type.code == e_withdraw.code:
                rec.name = camt053_get_unified_str(rec.detail_set.all(), 'creditor_name')
            elif rec.type.code == e_deposit.code:
                rec.name = camt053_get_unified_str(rec.detail_set.all(), 'debtor_name')
        if not rec.recipient_account_number:
            rec.recipient_account_number = camt053_get_unified_str(rec.detail_set.all(), 'creditor_account')
        if not rec.remittance_info:
            rec.remittance_info = camt053_get_unified_str(StatementRecordRemittanceInfo.objects.all().filter(detail__record=rec), 'reference')
        if not rec.paid_date:
            paid_date = camt053_get_unified_val(rec.detail_set.all(), 'paid_date', default=None)
            if paid_date:
                assert isinstance(paid_date, datetime)
                rec.paid_date = paid_date.date()

        rec.full_clean()
        rec.save()

    #pylint: enable=too-many-locals,too-many-statements,too-many-branches
    return stm



# Generated by Django 2.2.3 on 2019-11-30 20:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0054_wsediconnection_environment'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='wsediconnection',
            name='customer',
        ),
        migrations.AddField(
            model_name='wsediconnection',
            name='name',
            field=models.CharField(default='', max_length=64, verbose_name='name'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.3 on 2019-11-27 01:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0037_auto_20191127_0132'),
    ]

    operations = [
        migrations.AlterField(
            model_name='wsediconnection',
            name='signing_cert',
            field=models.FileField(blank=True, upload_to='certs', verbose_name='signing cert'),
        ),
    ]



#pylint: disable=logging-format-interpolation,too-many-branches
import logging
import os
import traceback
from django.core.management import CommandParser
from jbank.models import Payout, PAYOUT_ERROR, PAYOUT_WAITING_PROCESSING, PayoutStatus, PAYOUT_WAITING_UPLOAD, \
    WsEdiConnection
from jbank.sepa import Pain001, PAIN001_REMITTANCE_INFO_MSG, PAIN001_REMITTANCE_INFO_OCR_ISO, \
    PAIN001_REMITTANCE_INFO_OCR
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
        Generates pain.001.001.03 compatible SEPA payment files from pending Payout objects.
        By default generates files of Payouts in WAITING_PROCESSING state.
        """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('dir', type=str)
        parser.add_argument('--payout', type=int)
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--ws', type=int)
        parser.add_argument('--suffix', type=str, default='XL')

    def do(self, *args, **options):
        target_dir = options['dir']
        if options['verbose']:
            logger.info('Writing pain.001 files to {}'.format(target_dir))

        payouts = Payout.objects.all()
        if options['payout']:
            payouts = Payout.objects.filter(id=options['payout'])
        else:
            payouts = payouts.filter(state=PAYOUT_WAITING_PROCESSING)
        if options['ws']:
            ws = WsEdiConnection.objects.get(id=options['ws'])
            assert isinstance(ws, WsEdiConnection)
            if ws and not ws.enabled:
                logger.info('WS connection %s not enabled, exiting', ws)
                return
            payouts = payouts.filter(connection=ws)

        for p in list(payouts):
            assert isinstance(p, Payout)
            try:
                if options['verbose']:
                    print(p)

                if not p.msg_id:
                    p.generate_msg_id()
                if not p.file_name:
                    p.file_name = p.msg_id + '.' + options['suffix']
                    p.save(update_fields=['file_name'])

                pain001 = Pain001(p.msg_id, p.payer.name, p.payer.account_number, p.payer.bic, p.payer.org_id, p.payer.address_lines, p.payer.country_code)
                if p.messages:
                    remittance_info = p.messages
                    remittance_info_type = PAIN001_REMITTANCE_INFO_MSG
                else:
                    remittance_info = p.reference
                    remittance_info_type = PAIN001_REMITTANCE_INFO_OCR_ISO if remittance_info[:2] == 'RF' else PAIN001_REMITTANCE_INFO_OCR
                pain001.add_payment(p.msg_id, p.recipient.name, p.recipient.account_number, p.recipient.bic,
                                    p.amount, remittance_info, remittance_info_type, p.due_date)

                p.full_path = full_path = os.path.join(target_dir, p.file_name)
                if options['verbose']:
                    print(p, 'written to', full_path)
                pain001.render_to_file(full_path)
                logger.info('{} generated'.format(full_path))
                p.state = PAYOUT_WAITING_UPLOAD
                p.save(update_fields=['full_path', 'state'])

                PayoutStatus.objects.create(payout=p, file_name=p.file_name, msg_id=p.msg_id, status_reason="File generation OK")
            except Exception as e:
                short_err = 'File generation failed: ' + str(e)
                long_err = "File generation failed ({}): ".format(p.file_name) + traceback.format_exc()
                logger.error(long_err)
                p.state = PAYOUT_ERROR
                p.save(update_fields=['state'])
                PayoutStatus.objects.create(payout=p, file_name=p.file_name, msg_id=p.msg_id, status_reason=short_err[:255])



# Generated by Django 2.2.3 on 2019-11-30 20:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0052_auto_20191130_1927'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='target_identifier',
            field=models.CharField(default=1, max_length=32, verbose_name='target identifier'),
            preserve_default=False,
        ),
    ]



import logging
from django.core.management import CommandParser
from jutil.command import SafeCommand
from jutil.format import format_xml_bytes, format_xml
from jbank.helpers import validate_xml
from jbank.models import WsEdiConnection


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Makes test application request'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--ws', type=int, default=1)
        parser.add_argument('--xsd', type=str)
        parser.add_argument('--command', type=str, default='DownloadFileList')
        parser.add_argument('--file', type=str)

    def do(self, *args, **options):
        ws = WsEdiConnection.objects.get(id=options['ws'])
        if options['file']:
            content = open(options['file'], 'rb').read()
        else:
            content = ws.get_application_request(options['command']).encode()
        print('------------------------------------------------- Application request')
        print(format_xml_bytes(content).decode())
        if options['xsd']:
            validate_xml(content, options['xsd'])
        print('------------------------------------------------- Signed request')
        print(format_xml(ws.sign_application_request(content.decode())))



import logging
from pprint import pprint

from django.core.management import CommandParser
from jutil.command import SafeCommand
from cryptography import x509
from cryptography.hazmat.backends import default_backend


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Parses x509 cert'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('pem', type=str)

    def do(self, *args, **options):
        pem_data = open(options['pem'], 'rb').read()
        cert = x509.load_pem_x509_certificate(pem_data, default_backend())
        pprint(cert)



# Generated by Django 2.2.3 on 2019-11-30 19:23

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0050_remove_wsedisoapcall_payout'),
    ]

    operations = [
        migrations.AddField(
            model_name='payout',
            name='connection',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='jbank.WsEdiConnection'),
        ),
    ]



import logging
from datetime import datetime, timedelta
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from django.core.management import CommandParser
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
        Generates RSA private key and x509 certificate in .pem format (for testing)
        """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--key-file', type=str, default='key.pem')
        parser.add_argument('--cert-file', type=str, default='cert.pem')
        parser.add_argument('--country', type=str, default='US')
        parser.add_argument('--state', type=str, default='TX')
        parser.add_argument('--locality', type=str, default='Dallas')
        parser.add_argument('--org-name', type=str, default='Kajala Group')
        parser.add_argument('--common-name', type=str, default='kajala.com')

    def do(self, *args, **options):
        # Generate our key
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

        # Write to disk unencrypted
        with open(options['key_file'], "wb") as f:
            f.write(key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            ))
            print('{} written'.format(f.name))

        # Various details about who we are. For a self-signed certificate the
        # subject and issuer are always the same.
        subject = issuer = x509.Name([
            x509.NameAttribute(x509.NameOID.COUNTRY_NAME, options['country']),
            x509.NameAttribute(x509.NameOID.STATE_OR_PROVINCE_NAME, options['state']),
            x509.NameAttribute(x509.NameOID.LOCALITY_NAME, options['locality']),
            x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, options['org_name']),
            x509.NameAttribute(x509.NameOID.COMMON_NAME, options['common_name']),
        ])
        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.utcnow()
        ).not_valid_after(
            # Our certificate will be valid for 10 days
            datetime.utcnow() + timedelta(days=10)
        ).add_extension(
            x509.SubjectAlternativeName([x509.DNSName("localhost")]),
            critical=False,
        ).sign(key, hashes.SHA256(), default_backend())

        # Write our certificate out to disk.
        with open(options['cert_file'], "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))
            print('{} written'.format(f.name))



#pylint: disable=logging-format-interpolation,too-many-locals,too-many-branches
import base64
import logging
import os
import pytz
from django.core.management import CommandParser
from django.utils.timezone import now
from jutil.xml import xml_to_dict
from jbank.helpers import process_pain002_file_content, parse_start_and_end_date
from jbank.models import WsEdiConnection
from jbank.wsedi import wsedi_get, wsedi_execute
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
        Download Finnish bank files
        """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('path', type=str)
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--overwrite', action='store_true')
        parser.add_argument('--file-type', type=str, help='E.g. TO, SVM, XP, NDCORPAYL, pain.002.001.03')
        parser.add_argument('--status', type=str, default='', help='E.g. DLD, NEW')
        parser.add_argument('--file-reference', type=str, help='Download single file based on file reference')
        parser.add_argument('--list-only', action='store_true')
        parser.add_argument('--process-pain002', action='store_true')
        parser.add_argument('--start-date', type=str)
        parser.add_argument('--end-date', type=str)
        parser.add_argument('--ws', type=int)

    def do(self, *args, **options):  # pylint: disable=too-many-statements
        ws = WsEdiConnection.objects.get(id=options['ws']) if options['ws'] else None
        assert ws is None or isinstance(ws, WsEdiConnection)
        if ws and not ws.enabled:
            logger.info('WS connection %s not enabled, exiting', ws)
            return

        start_date, end_date = parse_start_and_end_date(pytz.timezone('Europe/Helsinki'), **options)
        path = options['path']
        command = 'DownloadFileList'
        time_now = now()
        file_reference = options['file_reference']
        if file_reference:
            command = 'DownloadFile'
        status = options['status']
        file_type = options['file_type']
        if command == 'DownloadFileList' and not file_type:
            print('--file-type required (e.g. TO, SVM, XP, NDCORPAYL, pain.002.001.03)')
            return
        if ws:
            content = wsedi_execute(ws, command=command, file_type=file_type, status=status, start_date=start_date, end_date=end_date,
                                    file_reference=file_reference, verbose=options['verbose'])
            data = xml_to_dict(content, array_tags=['FileDescriptor'])
        else:
            res = wsedi_get(command=command, file_type=file_type, status=status, file_reference=file_reference, verbose=options['verbose'])
            data = res.json()
            # "FileDescriptors": {
            #     "FileDescriptor": [
            #         {
            #             "FileReference": "535283541",
            #             "TargetId": "NONE",
            #             "UserFilename": "STOL001.FMV80KT2.WEBSER.PS",
            #             "ParentFileReference": "1218",
            #             "FileType": "TO",
            #             "FileTimestamp": "2017-12-18T20:33:09.362+02:00",
            #             "Status": "DLD",
            #             "LastDownloadTimestamp": "2017-12-19T12:36:34.490+02:00",
            #             "ForwardedTimestamp": "2017-12-18T20:33:09.362+02:00",
            #             "Deletable": "false",
            #             "CustomerNumber": "06720106",
            #             "Modifier": "06720106",
            #             "ModifiedTimestamp": "2017-12-19T12:36:34.490+02:00",
            #             "SourceId": "A",
            #             "Environment": "PRODUCTION"
            #         },
            #         ...

        if command == 'DownloadFileList':
            if 'FileDescriptors' in data and data['FileDescriptors'] is not None and 'FileDescriptor' in data['FileDescriptors']:
                for fd in data['FileDescriptors']['FileDescriptor']:
                    file_reference = fd['FileReference']
                    file_type = fd['FileType']
                    file_basename = file_reference + '.' + file_type
                    file_path = os.path.join(path, file_basename)
                    if options['list_only']:
                        print('{file_reference} ({file_type}/{status}): {user_filename} ({timestamp})'.format(
                            file_reference=file_reference, file_type=file_type, status=fd.get('Status'),
                            user_filename=fd.get('UserFilename'), timestamp=fd.get('FileTimestamp')))
                        continue
                    if options['overwrite'] or not os.path.isfile(file_path):
                        command = 'DownloadFile'
                        if ws:
                            content = wsedi_execute(ws, command=command, file_type=file_type, status='', file_reference=file_reference,
                                                    verbose=options['verbose'])
                            file_data = xml_to_dict(content)
                        else:
                            res = wsedi_get(command=command, file_type=file_type, status='', file_reference=file_reference, verbose=options['verbose'])
                            file_data = res.json()
                        if 'Content' not in file_data:
                            logger.error('WS-EDI {} HTTP {} but Content block missing: {}'.format(command, res.status_code, file_data))
                            raise Exception("WS-EDI {} HTTP {} but Content block missing".format(command, res.status_code))
                        bcontent = base64.b64decode(file_data['Content'])
                        with open(file_path, 'wb') as fp:
                            fp.write(bcontent)
                        logger.info('Wrote file {}'.format(file_path))

                        # process selected files immediately
                        if options['process_pain002'] and file_type in ['XP', 'pain.002.001.03', 'NDCORPAYL']:
                            process_pain002_file_content(bcontent, file_path, created=time_now)
                    else:
                        print('Skipping old file {}'.format(file_path))
            else:
                print('Empty file list downloaded')
        elif command == 'DownloadFile':
            bcontent = base64.b64decode(data['Content'])
            file_path = os.path.join(path, file_reference)
            if options['overwrite'] or not os.path.isfile(file_path):
                with open(file_path, 'wb') as fp:
                    fp.write(bcontent)
                logger.info('Wrote file {}'.format(file_path))
            else:
                print('Skipping old file {}'.format(file_path))



#pylint: disable=logging-format-interpolation
import logging
from django.core.management import CommandParser
from jbank.models import StatementFile, StatementRecord
from jbank.parsers import parse_tiliote_statements_from_file
from jutil.command import SafeCommand


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Re-parses old bank statement .TO (tiliote) files. Used for adding missing fields.'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--file', type=str)

    def do(self, *args, **options):
        logger.info('Re-parsing TO files to update fields')
        qs = StatementFile.objects.all()
        if options['file']:
            qs = qs.filter(file=options['file'])
        for file in qs.order_by('id'):
            assert isinstance(file, StatementFile)
            logger.info('Processing {} BEGIN'.format(file))
            statements = parse_tiliote_statements_from_file(file.full_path)
            for data in statements:
                for e in data['records']:
                    # check missing line_number
                    e2 = StatementRecord.objects.filter(statement__file=file, line_number=0, record_number=e['record_number'],
                                                        archive_identifier=e['archive_identifier'], record_date=e['record_date'],
                                                        value_date=e['value_date'], paid_date=e['paid_date'],
                                                        entry_type=e['entry_type'], record_code=e['record_code'],
                                                        record_description=e['record_description'], receipt_code=e['receipt_code'],
                                                        delivery_method=e['delivery_method'], name=e['name'],
                                                        name_source=e['name_source'], recipient_account_number=e['recipient_account_number'],
                                                        recipient_account_number_changed=e['recipient_account_number_changed'],
                                                        remittance_info=e['remittance_info']).first()
                    if e2:
                        e2.line_number = e['line_number']
                        e2.save()
                        logger.info('Updated {} line number to {}'.format(e2, e2.line_number))
            logger.info('Processing {} END'.format(file))



import logging
from decimal import Decimal
from django.core.management import CommandParser
from jutil.command import SafeCommand
from jbank.helpers import make_msg_id
from jbank.models import Payout, PayoutParty

logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Makes test payment'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--payer-id', type=int, default=1)
        parser.add_argument('--recipient-id', type=int, default=2)
        parser.add_argument('--messages', type=str, default='test payment')
        parser.add_argument('--amount', type=Decimal, default=Decimal('1.23'))
        parser.add_argument('--ws', type=int, default=1)

    def do(self, *args, **options):
        payer = PayoutParty.objects.get(id=options['payer_id'])
        p = Payout(account=payer.payouts_account, payer=payer, recipient_id=options['recipient_id'], messages=options['messages'], msg_id=make_msg_id(), amount=options['amount'], connection_id=options['ws'])
        p.full_clean()
        p.save()
        print('{} created'.format(p))



# Generated by Django 2.2.3 on 2019-11-27 05:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0040_wsediconnection_signing_key_file'),
    ]

    operations = [
        migrations.AddField(
            model_name='wsediconnection',
            name='encryption_cert_file',
            field=models.FileField(blank=True, upload_to='certs', verbose_name='encryption certificate file'),
        ),
        migrations.AddField(
            model_name='wsediconnection',
            name='encryption_key_file',
            field=models.FileField(blank=True, upload_to='certs', verbose_name='encryption key file'),
        ),
    ]
